/*
 * Copyright  2003 - 2012, SAS Institute Inc., Cary, NC, USA. All Rights Reserved
 * 
 * This software is protected by copyright laws and international treaties.
 * 
 * U.S. GOVERNMENT RESTRICTED RIGHTS
 * Use, duplication, or disclosure of this software and related documentation by the U.S. government is 
 * subject to the Agreement with SAS Institute and the restrictions set forth in FAR 52.227-19, 
 * Commercial Computer Software - Restricted Rights (June 1987). 
 * 
 */

/*
 * This script migrates attachments from the MRM database to the SAS Content Server.
 * It must be run by a MRM Administrator.
 * 
 * 1. Upload this script to the SAS Content Server at the following path:
 *    
 *   For 7.4    ->  sasdav/Products/SASRiskGovernanceFramework/RiskGovFrwkMid-Tier/mrm/Scripts/migrateattachments.groovy
 *   Before 7.4 ->  sasdav/Products/SASModelRisk/ModelRiskMgtMid-Tier/Scripts/migrateattachments.groovy
 * 
 * 2. Log into MRM as a MRM Administrator
 * 
 * 3. Access this script via the web interface at:
 *    For 7.4    -> http://<server>:<port>/SASRiskGovernanceFramework/migrateattachments.groovy?rgfContentId=mrm
 *    Before 7.4 -> http://<server>:<port>/SASModelRisk/migrateattachments.groovy
 *     
 * 4. After running the script, you may delete it from the SAS Content Server if you wish
 */

import java.sql.Connection
import java.sql.ResultSet
import java.sql.SQLException

import org.apache.commons.io.IOUtils
import org.apache.commons.logging.LogFactory;


import com.sas.fsd.shared.services.content.DAVContentServiceException
import com.sas.fsd.shared.services.content.DAVResourceContent
import com.sas.fsd.shared.services.content.WriteDAVResourcesTransaction
import com.sas.fsd.shared.util.StringUtil
import com.sas.oprisk.framework.server.OpRiskException
import com.sas.oprisk.framework.server.objects.TaggedPointerHandle
import com.sas.oprisk.framework.server.objects.BusinessObject.ClassObject
import com.sas.oprisk.framework.server.objects.BusinessObject
import com.sas.oprisk.framework.server.persistence.PersistenceException
import com.sas.oprisk.monitor.web.MonitorContext
import com.sas.oprisk.monitor.web.MonitorServlet
import com.sas.oprisk.server.Attachment
import com.sas.oprisk.server.BusinessObjectTypeInfo
import com.sas.oprisk.server.behavior.PrimaryBusinessObjectRef
import com.sas.oprisk.server.content.Content
import com.sas.oprisk.server.content.ContentFactory;
import com.sas.oprisk.server.content.ServerContentService
import com.sas.oprisk.server.services.ServiceLocator;
import com.sas.oprisk.server.util.BusinessObjectUtils




///////////////////////////////////////////////////////////////////////////////
// Main handler
//
def handle() {
   setup();
   try {
      if (verifyPermissions() && verifyVersion()) {
         
         action = StringUtil.trimToNull(request.getParameter("action"));
         if ("migrate".equals(action) && handleMigrate()) {
            return;
         }
         else if ("purge".equals(action) && handlePurge()) {
            return;
         }
         else {
            handleHome();
         }
      }
   }
   finally {
      cleanup();
   }
}



def setup() {
   psession=MonitorContext.getPersistenceSession (request);
   psession.rollback(); // Make sure the DB connection is not stuck in a transaction

   adminId = currentUser.userId
   
   contentService = ServiceLocator.getServerContentService();
   contentFactory = new ContentFactory(contentService);
   
   logger = LogFactory.getLog( MonitorServlet.class );
}
def cleanup() {
   psession.rollback(); // Make sure the DB connection is not stuck in a transaction
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission and version checks
//
def verifyPermissions() {
   if (currentUser.isSuperUser(psession))
      return true;

   printError("This tool may only be used by MRM Administrators");
   return false;
}
def verifyVersion() {
   def methodNames = Content.class.declaredMethods.collect({method -> method.name})

   if (!methodNames.contains("updateAttachment"))
   {
      // A missing method means we're not on 51M1 or higher
      printError("This utility requires SAS MRM version 6.1 or higher")
      return false;
   }

   // A missing method means we don't support groupNumber (added in 5.1M3)
   groupsSupported = methodNames.contains("setGroupNo")

   // A missing method means we don't support pointInTime (added in 5.1M4)
   methodNames = BusinessObject.class.declaredMethods.collect({method -> method.name})
   pointInTimeSupported = methodNames.contains("setPointInTime")
   
   return true;
}



//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Utility methods
//
def printWarning(msg) {
   out.println("""
<div class="warningHeader">
   <b>WARNING</b><br>
   $msg
</div>
""")
}
def printError(msg, throwable=null) {
   out.println("""
<div class="errorHeader">
   <b>ERROR</b><br>
   $msg
""")
   if (throwable != null) {
      out.println("""<a href="javascript:;" onclick="this.style.display='none';this.nextSibling.style.display='block';">Show details...</a><div style='display:none;'><pre>""")
      throwable.printStackTrace(out)
      out.println("""</pre></div>""")
   }
   out.println("""
</div>
""")
}

def updateProgress(progress) {
   out.println("<script>updateProgress($progress);</script>")
   out.flush();
}

ERROR = "error"
WARNING = "warning"
INFO = "info"
DEBUG = "debug"
def updateStatus(status, level) {
   if (level == ERROR) 
      logger.error(status);
   else if (level == WARNING) 
      logger.warning(status);
   else 
      logger.info(status);
   
   String escapedStatus = StringUtil.escapeJSForHTML(status);
   out.println("<script>updateStatus('$escapedStatus','$level');</script>")
   out.flush();
}

def printHomeLink() {
   out.println("""<br><a href="?">Return to the migration tool</a>""")
}

def filesize (attrs) {
   labels = [' bytes','KB','MB','GB','TB']
   size = attrs.size
   label = labels.find {
      if( size < 1024 ) {
         true
      }
      else {
         size /= 1024
         false
      }
   }
   return "${new java.text.DecimalFormat( attrs.format ?: '0.##' ).format( size )}$label"
}

def createBusinessObjectRef(TaggedPointerHandle tpHandle)
{
   String pointerTag = tpHandle.getPointerTag();
   ClassObject classObject = BusinessObjectUtils.getClassObjectForPointerTag(pointerTag);
   String businessObjectTypeNm = classObject.getBusinessObjectTypeName();
   Long businessObjectRk = tpHandle.getPointer();
   return PrimaryBusinessObjectRef.createRef(businessObjectTypeNm, businessObjectRk);
}

def getTypeFilter() {
   String type = StringUtil.trimToNull(request.getParameter("type"))
   if ("ALL".equals(type)) {
      return null
   }
   else {
      BusinessObjectTypeInfo.getByBusinessObjectTypeName(type)
      return type;
   }
}
def getGroupFilter() {
   return StringUtil.trimToNull(request.getParameter("group"))
}










//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Migration handling
//
def handleMigrate() {
   // Do migrate
   start = System.currentTimeMillis()
   try {
      updateProgress(0)
      visitAttachments(getTypeFilter(), getGroupFilter(), { attachment, number, total ->
         try {
            if (attachment != null) {
               migrate(attachment);
            }
         }
         finally
         {
            updateProgress(Math.ceil(number * 100.0/total))
         }
      })
   }
   catch(Exception e){
      String message = "An error occurred while migrating:"
      updateStatus(message, ERROR)
      updateStatus("\t" + e.getLocalizedMessage(), ERROR)
   }
   finally {
      end = System.currentTimeMillis()
      updateProgress(100)
      updateStatus("Migration complete", INFO)
      updateStatus("Total time: " + ((end-start)/1000.0) + " seconds", DEBUG)
   }
   printHomeLink()
   return true;
}

def migrate(Attachment attachment) {
   PrimaryBusinessObjectRef boRef = createBusinessObjectRef(attachment.getAttachedToTaggedPointerHandle());
   Content content = getDAVResourceForAttachment(boRef, attachment.getFileNm());
   if (content == null)
   {
      // Error
      throw new RuntimeException("Could not create DAV entry for " + attachment);
   }
   else
   {
      String path = boRef.getBusinessObjectTypeNm() + "/" + boRef.getBusinessObjectRk() + "/" + attachment.getFileNm();
      updateStatus(path, INFO);

      String message = "\t" + (content.isNew() ? "Creating new DAV item" : "DAV item already exists");
      updateStatus(message, DEBUG);

      copyContentToContentServer(content, attachment);
      checkBinaryContent(boRef, attachment);
      changeDatabaseRowType(content, attachment);
   }
}

def getDAVResourceForAttachment(PrimaryBusinessObjectRef<?> boRef, String attachmentName) {
   Content content = null;
   try {
      content = (Content) contentService.readDAVResource(boRef, attachmentName, adminId, null, contentFactory);
   }
   catch (Exception e) {
      // Might not exist, so ignore errors here
   }
   if (content == null) {
      content = (Content) contentService.createDAVResource(boRef, adminId, null, contentFactory);
      content.setFileName(attachmentName);
      content.setRevisionCd("PUBLISHED");
   }
   return content;
}

def copyContentToContentServer(Content content, Attachment attachment) throws SQLException, IOException, OpRiskException, DAVContentServiceException
{
   // Ensure binary content matches
   byte[] dbContent = attachment.getContent(psession);

   InputStream inputStream = content.getInputStream(contentService);
   byte[] davContent = inputStream == null ? null : IOUtils.toByteArray(inputStream);
   if (!Arrays.equals(dbContent, davContent))
   {
      // Persist content to DAV
      content.setContent(new DAVResourceContent(){
            public String getFileName() { return attachment.getFileNm(); }
            public long getFileSize() { return dbContent.length; }
            public String getContentType() { return attachment.getContentType(); }
            public InputStream getInputStream() { return new ByteArrayInputStream(dbContent); }
      });
   }

   content.setBusinessObjectRef(createBusinessObjectRef(attachment.getAttachedToTaggedPointerHandle()));
   content.setFileName(attachment.getFileNm());
   content.setDescription(attachment.getAttachmentNm());
   content.setAttachmentRk(attachment.getAttachmentRk());

   if (groupsSupported)
      content.setGroupNo(attachment.getGroupNo());

   if (content.isDirty())
   {
      if (content.isNew())
         content.setAttribute("{DAV:}creationdate", attachment.getValidFrom())
      content.setAttribute("{DAV:}getlastmodified", attachment.getValidFrom())
	  
      contentService.processTransaction(new WriteDAVResourcesTransaction(content));
      String message = "\tUpdated DAV content";
      updateStatus(message, DEBUG);
   }
   else {
      String message = "\tDAV content already up-to-date";
      updateStatus(message, DEBUG);
   }
}

def checkBinaryContent(PrimaryBusinessObjectRef<?> boRef, Attachment attachment) throws PersistenceException, IOException
{
   // Refetch to verify against what's on the server
   Content content = getDAVResourceForAttachment(boRef, attachment.getFileNm());

   // Ensure binary content matches
   byte[] dbContent = attachment.getContent(psession);

   InputStream inputStream = content.getInputStream((ServerContentService) contentService);
   byte[] davContent = inputStream == null ? null : IOUtils.toByteArray(inputStream);
   if (content.getFileSize() != dbContent.length)
      throw new RuntimeException("Binary content length didn't match for " + content.getPath().getFullPath());
   if (!Arrays.equals(dbContent, davContent))
      throw new RuntimeException("Binary content didn't match for " + content.getPath().getFullPath());

   if (!createBusinessObjectRef(attachment.getAttachedToTaggedPointerHandle()).equals(content.getBusinessObjectRef()))
      throw new RuntimeException("Business object ref didn't match for " + content.getPath().getFullPath());

   if (!attachment.getFileNm().equals(content.getFileName()))
      throw new RuntimeException("Attachment name didn't match description for " + content.getPath().getFullPath());

   if (!attachment.getAttachmentNm().equals(content.getDescription()))
      throw new RuntimeException("Content attachment name didn't match for " + content.getPath().getFullPath());

   if (!attachment.getAttachmentRk().equals(content.getAttachmentRk()))
      throw new RuntimeException("Content attachment RK didn't match for " + content.getPath().getFullPath());

   if (groupsSupported && attachment.getGroupNo() != content.getGroupNo())
      throw new RuntimeException("Content attachment groupNumber didn't match for " + content.getPath().getFullPath());

   String message = "\tVerified DAV content matches";
   updateStatus(message, DEBUG);
}

def changeDatabaseRowType(Content content, Attachment attachment)
{
   content.updateAttachment(attachment, psession);
   attachment.save("Migrate to SAS Content Server", psession);
   psession.commit();
   updateStatus("\tUpdated DB row type to CONTENT", DEBUG);
}


def visitAttachments(type, group, visitor) {
   int i=1
   Set<Long> rks = getAttachmentRks(type, group);
   long count = rks.size()
   for (long rk : rks)
   {
      Attachment attachment = null
      try
      {
         attachment = Attachment.object.fetchIfExists(rk, psession)
      }
      catch(Exception e) {
         String message = "Error fetching attachment " + rk
         updateStatus(message, ERROR)
      }
      
      try
      {
         visitor(attachment, i, count)
      }
      catch (Exception e)
      {
         String message = "Error processing " + attachment.getFileNm()
         updateStatus(message, ERROR)
         updateStatus("\t" + e.getLocalizedMessage(), ERROR)
      }
      i++;
   }
}

def getAttachmentRks(type, group) {
   Connection conn = psession.asSqlPersistenceSession().getConnection();
   
   conditions = []

   conditions += "attachment_type_cd = 'FIL'"

   if (group == null || group.equals("") || group.equals("*")) {
      // No group restrictions
   }
   else if (group.equals("null")) {
         // Require an empty group
         conditions += "group_no IS NULL"
   }
   else {
         // Require a specific group
         conditions += "group_no = ${Integer.parseInt(group)}"
   }
   
   if (type != null) {
      conditions += "business_object_nm = '$type'"
   }

   ResultSet rs = conn.createStatement().executeQuery("""
SELECT 
      attachment_rk 
FROM
      attachment_l 
WHERE
      ${conditions.join(" AND ")}
ORDER BY
      business_object_nm, 
      group_no
""");

   Set<Long> result = new LinkedHashSet();
   try{
      while(rs.next()) {
         result.add(rs.getLong(1))
      }
   }
   finally {
      //rs.close();
   }
   
   return result
}










//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Purge handling
//
def handlePurge() {
   // Do purge
   start = System.currentTimeMillis()
   archive = Boolean.parseBoolean(request.getParameter("archive"));
   
   try {
      purge(archive);
   }
   catch(Exception e){
      String message = "An error occurred while removing migrated data:"
      updateStatus(message, ERROR)
      updateStatus("\t" + e.getLocalizedMessage(), ERROR)
   }
   finally {
      end = System.currentTimeMillis()
      updateStatus("Removal complete", INFO)
      updateStatus("Total time: " + ((end-start)/1000.0) + " seconds", DEBUG)
   }
   
   printHomeLink()
   return true;
}
def purge(archive) {
   try {
      table = archive ? "ATTACHMENT":"ATTACHMENT_L"
      updateStatus("Removing migrated content from $table table", INFO)
      
      Connection conn = psession.asSqlPersistenceSession().getConnection();
      String condition = getPurgeCondition(archive)
      count = 0
      if (archive) {
         count = conn.createStatement().executeUpdate("UPDATE attachment SET content=NULL WHERE $condition");
      }
      else {
         count = conn.createStatement().executeUpdate("UPDATE attachment_l SET content=NULL WHERE $condition");
      }
      psession.commit()
      
      updateStatus("Updated $count ${count == 1 ? 'row':'rows'}", INFO)
   }
   finally {
      updateProgress(100)
   }
}










///////////////////////////////////////////////////////////////////////////////
// Home handler
//
def handleHome() {
   printWarning("""
This tool will migrate attachment content from the database to the SAS Content Server.<br>
<b><u>This process is not reversible. Ensure you have a backup of the MRM database and SAS Content Server content before proceeding.</b></u> 
""")
   
   out.println("""
      <hr>
      <h2>1. Prepare modified screen definitions</h2>
      Screen definitions for the parent business objects of the attachments you want to migrate must be updated to use the ContentManagement component:
      <pre>
      &lt;field type="component" component-name="ContentManagement"&gt;
         &lt;label&gt;...&lt;/label&gt;
      &lt;/field&gt;</pre>
""")
   
   if (!groupsSupported || !pointInTimeSupported) {
      def unsupportedFeatures = []
      if (!groupsSupported)
         unsupportedFeatures << "groupNumber"
      if (!pointInTimeSupported)
         unsupportedFeatures << "pointInTime"
      printWarning("""
      The ContentManagement component does not support ${unsupportedFeatures.join(" or ")} functionality.
      <br>
      If an existing screen definition requires that functionality, do not migrate the attachments for that business object type.
""")
   }

   out.println("<hr><h2>2. Move attachments from the MRM database to the SAS Content Server</h2>")
   printMigrateForm();

   out.println("""
      <hr>
      <h2>3. Activate modified screen definitions</h2>
      Activate all the screen definitions modified in step #1.
      <br>
      Migrated attachments will appear in the ContentManagement component.
""")

   out.println("<hr>")
   out.println("<h2>4. Remove migrated content from the MRM database (ATTACHMENT_L table)</h2>")
   printPurgeForm(false);

   out.println("<hr>")
   out.println("<h2>5. Remove migrated content from the MRM database archive (ATTACHMENT table)</h2>")
   printPurgeForm(true);
}





def getAttachmentCounts() {
   Connection conn = psession.asSqlPersistenceSession().getConnection();
   ResultSet rs = conn.createStatement().executeQuery("""
SELECT 
      business_object_nm, 
      group_no, 
      count(attachment_rk), 
      sum(length(content)) 
FROM
      attachment_l 
WHERE
      attachment_type_cd = 'FIL'
GROUP BY
      business_object_nm, 
      group_no 
ORDER BY
      business_object_nm, 
      group_no
""");

   all = [
      type:"ALL", 
      displayType:"<b>All attachments</b>",
      group:"*",
      displayGroup:"&nbsp;",
      count:0,
      size:0
   ] 
   counts = []
   
   try{
      while(rs.next()) {
         String businessObjectNm = rs.getString(1)
         String displayBusinessObjectNm = businessObjectNm
         
         String type = rs.getString(1)
         String displayType = type;
         try {
            displayType = BusinessObjectUtils.getClassObjectForPointerTag(type).getBusinessObjectInfo().getDisplayNameResource();
         }
         catch(Exception e) {            
            e.printStackTrace();
         }
         
         String group = rs.getString(2)
         
         long count = rs.getLong(3)
         all.count += count
         
         long size = rs.getLong(4)
         all.size += size
         
         counts += [
            type: type,
            displayType: displayType,
            
            group: group ?: "null",
            displayGroup: group ?: "&nbsp;",
            warnGroup: !groupsSupported && group != null,
            
            count: count,
            
            size: size,
            displaySize: filesize([size:size])
         ]
         
         if (!groupsSupported && group != null)
            all.warnGroup = true
      }
   }
   finally {
      //rs.close();
   }
   
   if (all.size > 0) {
      all.displaySize = filesize([size:all.size])
      counts.add(0, all)
   }
   
   counts = counts.sort({ a, b ->
      if (a.type == "ALL")
         1
      else if (b.type == "ALL")
         -1
      else
         a.displayType <=> b.displayType
   })
   
   return counts
}

def printMigrateForm() {
   counts = getAttachmentCounts()
   if (counts.size == 0) {
      out.println("No unmigrated attachments found")
   }
   else {
      groupWarning = """<span class="warningIcon" title="Group numbers are not supported by SAS Content Server attachments, and will be lost when migrating">!</span>"""
      out.println("""
<table border="1">
   <tr>
      <th>Parent Business Object</th>
      <th>Attachment Count</th>
      <th>Group Number</th>
      <th>Total File Size</th>
   </tr>""");

      counts.each { count ->
         onsubmit = ""
         if (count.warnGroup)
            onsubmit = """onsubmit="return confirm('WARNING: Migrating these attachments will discard the existing groupNumber information. Continue anyway?');" """

         out.println("""
   <tr>
      <td class="type">${count.displayType}</td>
      <td class="count">${count.count}</td>
      <td class="group">${count.displayGroup} ${count.warnGroup ? groupWarning:""}</td>
      <td class="size">${count.displaySize}</td>
      <td class="submit">
         <form method="post" ${onsubmit}>
            <input type="hidden" name="action" value="migrate">
            <input type="hidden" name="type" value="${count.type}">
            <input type="hidden" name="group" value="${count.group}">
            <input type="submit" value="Migrate ${count.count} ${count.count == 1 ? 'attachment':'attachments'}">
         </form>
      </td>
   </tr>""")
      }
      
      out.println("</table>");
   }
}





def getPurgeCondition(archive) {
   if (archive)
      return "attachment_rk IN (SELECT attachment_rk FROM attachment WHERE attachment_type_cd='CON') AND content IS NOT NULL"
   else 
      return "attachment_type_cd = 'CON' AND content IS NOT NULL"
}
def getPurgeCounts(archive) {
   Connection conn = psession.asSqlPersistenceSession().getConnection();
   ResultSet rs;
   String condition = getPurgeCondition(archive)
   if (archive) {
      rs = conn.createStatement().executeQuery("""
SELECT
      count(attachment_rk),
      sum(length(content))
FROM
      attachment
WHERE
      $condition
""");
   }
   else {
      rs = conn.createStatement().executeQuery("""
SELECT
      count(attachment_rk),
      sum(length(content))
FROM
      attachment_l
WHERE
      $condition
""");
   }

   retval = [count: 0, size: 0]
   try{
      if(rs.next()) {
         retval.count = rs.getLong(1)
         retval.size = rs.getLong(2)
         retval.displaySize = filesize([size:retval.size])
      }
   }
   finally {
      //rs.close();
   }
   return retval
}
def purge

def printPurgeForm(archive) {
   data = getPurgeCounts(archive)
   table = archive ? "ATTACHMENT":"ATTACHMENT_L"
   if (data.count == 0) {
      out.println("No migrated binary content found in the ${table} table")
   }
   else {
      out.println("""
<table border="1">
   <tr>
      <th>Row Count</th>
      <th>Total File Size</th>
   </tr>
   <tr>
      <td class="count">${data.count}</td>
      <td class="size">${data.displaySize}</td>
      <td class="submit">
         <form method="post">
            <input type="hidden" name="action" value="purge">
            <input type="hidden" name="archive" value="${archive}">
            <input type="submit" value="Remove binary content from ${data.count} ${data.count == 1 ? 'row':'rows'}">
         </form>
      </td>
   </tr>
</table>""");
   }
}










///////////////////////////////////////////////////////////////////////////////
// Main HTML document
//
out.println("""<!DOCTYPE html>
<html>
<head>
   <title>Migrate Attachments to SAS Content Server</title>
   <style>
      body, p, td, th { font-size: 10pt }
      body, p, td, th { font-family: sans-serif; }

      form { display: inline }
      input { display: block; font-size: .8em; width: 100%; padding: 0 3px;}

      table { border-collapse: collapse }
      th, td { padding: 3px }
      td.group, td.count, td.size { text-align: right; }

      .warningHeader {
         border-radius: 1em;
         padding: 1em;
         margin-bottom: 2em;

         background: #ffd65e;
         background: -moz-linear-gradient(top,  #ffd65e 0%, #febf04 100%);
         background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#ffd65e), color-stop(100%,#febf04));
         background: -webkit-linear-gradient(top,  #ffd65e 0%,#febf04 100%);
         background: -o-linear-gradient(top,  #ffd65e 0%,#febf04 100%);
         background: -ms-linear-gradient(top,  #ffd65e 0%,#febf04 100%);
         background: linear-gradient(top,  #ffd65e 0%,#febf04 100%);
         filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#ffd65e', endColorstr='#febf04',GradientType=0 );
      }
      .errorHeader {
         border-radius: 1em;
         padding: 1em;
         margin-bottom: 2em;

         background: #ff9e93;
         background: -moz-linear-gradient(top,  #ff9e93 0%, #fc7676 100%);
         background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#ff9e93), color-stop(100%,#fc7676));
         background: -webkit-linear-gradient(top,  #ff9e93 0%,#fc7676 100%);
         background: -o-linear-gradient(top,  #ff9e93 0%,#fc7676 100%);
         background: -ms-linear-gradient(top,  #ff9e93 0%,#fc7676 100%);
         background: linear-gradient(top,  #ff9e93 0%,#fc7676 100%);
      }

      .warningIcon {
         display: inline-block;
         width: 1.3em;
         border-radius: 1em;
         border: 1px solid grey;
         font-weight: bold;
         text-align: center;
         cursor: default;

         background: #fceabb;
         background: -moz-linear-gradient(top,  #fceabb 0%, #fccd4d 50%, #f8b500 51%, #fbdf93 100%);
         background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#fceabb), color-stop(50%,#fccd4d), color-stop(51%,#f8b500), color-stop(100%,#fbdf93));
         background: -webkit-linear-gradient(top,  #fceabb 0%,#fccd4d 50%,#f8b500 51%,#fbdf93 100%);
         background: -o-linear-gradient(top,  #fceabb 0%,#fccd4d 50%,#f8b500 51%,#fbdf93 100%);
         background: -ms-linear-gradient(top,  #fceabb 0%,#fccd4d 50%,#f8b500 51%,#fbdf93 100%);
         background: linear-gradient(top,  #fceabb 0%,#fccd4d 50%,#f8b500 51%,#fbdf93 100%);
         filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#fceabb', endColorstr='#fbdf93',GradientType=0 );
      }

      #progressContainer {
         display: none;
         width:0%;
         min-width: 30px;
         text-align: center;
         font-family: sans-serif;
         font-weight: bold;
         font-size: 12px;
         color: white;
         border: 1px solid #888;
         padding: 5px 0px;
         margin-bottom: 10px;
         margin-left: -1px;
         
         background: #bfd255;
         background: -moz-linear-gradient(top,  #bfd255 0%, #8eb92a 50%, #72aa00 51%, #9ecb2d 100%);
         background: -webkit-gradient(linear, left top, left bottom, color-stop(0%,#bfd255), color-stop(50%,#8eb92a), color-stop(51%,#72aa00), color-stop(100%,#9ecb2d));
         background: -webkit-linear-gradient(top,  #bfd255 0%,#8eb92a 50%,#72aa00 51%,#9ecb2d 100%);
         background: -o-linear-gradient(top,  #bfd255 0%,#8eb92a 50%,#72aa00 51%,#9ecb2d 100%);
         background: -ms-linear-gradient(top,  #bfd255 0%,#8eb92a 50%,#72aa00 51%,#9ecb2d 100%);
         background: linear-gradient(top,  #bfd255 0%,#8eb92a 50%,#72aa00 51%,#9ecb2d 100%);
         filter: progid:DXImageTransform.Microsoft.gradient( startColorstr='#bfd255', endColorstr='#9ecb2d',GradientType=0 );
      }
      #statusContainer {
         display: none;
         overflow:auto;
         height:500px;
         font-family:monospace;
         font-size: 11px;
         border: 1px solid #aaa;
         padding: 10px;
         white-space: pre;
      }
      #statusContainer .error   { color: red }
      #statusContainer .warning { color: orange }
      #statusContainer .debug   { color: #AAA }

      h2 { margin: 0 0 10px 0; padding: 0 }
      hr { margin: 20px 0; padding: 0 }
   </style>
</head>
<body>
   <div id="progressContainer">0%</div>
   <div id="statusContainer"></div>
   <script>
      var progress = document.getElementById('progressContainer');
      var container = document.getElementById('statusContainer');
      function updateProgress(percent) {
         progress.style.display = "block";
         progress.style.width = percent + '%';
         progress.innerHTML = percent + '%';
      }
      function updateStatus(msg, cssClass) {
         container.style.display = "block";
         var msgDiv = document.createElement('div');
         msgDiv.className = cssClass;
         msgDiv.appendChild(document.createTextNode(msg));
         container.appendChild(msgDiv);
         container.scrollTop = container.scrollHeight;
      }
   </script>
""")

try {
   handle();
}
catch(Throwable t) {
   printError("An error was encountered while running this utility", t)
   printHomeLink()
}

out.println("""
</body>
</html>
""")
